"""Module containing queue selection algorithms.


How to implement your own queue selection algorithm?

    class MyQueueSelector(SelectQueueBase):

        def get_queue(self):

            # Get a list of all queues defined in the YAML file that is
            # specified at QUEUE_CONFIG in settings file.
            #
            # You may use these two properties of a queue object to select
            # a queue:
            #
            # - name: string of queue name
            # - priority: integer of queue priority; larger value,
            #       higher priority
            queues = self.queue_info.get_queues()

            # Implement your algorithm here
            # ...

            # Eventually, return one of queue object from queues
            return queue
"""
from __future__ import absolute_import

import random

from six.moves import range


class SelectQueueBase(object):
    """Base class for selecting a queue.

    The only method that needs to be implemented:

    get_queue: it's called for each task processing cycle on task worker.
    """

    def __init__(self, queue_info):
        self.queue_info = queue_info

    def get_queue(self, *args, **kwargs):
        """Returns a TaskQueue object."""
        raise NotImplementedError('Base class cannot be used directly.')


class Random(SelectQueueBase):
    """Randomly selects a queue without considering priority."""

    def get_queue(self):
        queues = self.queue_info.get_queues()
        return random.choice(queues)


class Lottery(SelectQueueBase):
    """Use lottery scheduling algorithm to select a queue based on priority."""

    @staticmethod
    def _run_lottery(queues):
        """Draw lottery from a list of candidate queues.

        :param list[TaskQueue] queues: a list of candidate queues.

        :return: A TaskQueue object that wins lottery. If it fails (e.g.,
            invalid priority of queues), it returns None.
        :rtype: TaskQueue
        """
        tickets = {}
        total_tickets = 0
        for queue in queues:
            # Queue priority should be within 1 to 100.
            if queue.priority < 1 or queue.priority > 100:
                continue
            priority = queue.priority
            low = total_tickets
            total_tickets += priority
            high = total_tickets
            tickets[queue.name] = (low, high)

        # [0, total_tickets)
        try:
            number = random.randrange(0, total_tickets)
            for queue in queues:
                if number >= tickets[
                        queue.name][0] and number < tickets[queue.name][1]:
                    return queue
        except ValueError:
            return None

        # Something wrong happens
        return None

    def get_queue(self, *args, **kwargs):
        return self._run_lottery(self.queue_info.get_queues())


class HighestPriorityFirst(SelectQueueBase):
    """Highest priority first.

    Get the highest priority non-empty queue first.
    If all queues are empty, get the highest priority empty queue.
    """

    def get_queue(self, *args, **kwargs):
        queue = self.queue_info.get_highest_priority_queue_that_needs_work()
        if queue:
            return queue
        queues = self.queue_info.get_queues()
        queues.sort(key=lambda x: x.priority, reverse=True)
        return queues[0]


class HighestPriorityLottery(Lottery):
    """Highest priority first  + lottery.

    Get highest priority non-empty queue first.
    If all queues are empty, run lottery on empty queues.
    """

    def get_queue(self, *args, **kwargs):
        queue = self.queue_info.get_highest_priority_queue_that_needs_work()
        if queue:
            return queue

        return self._run_lottery(self.queue_info.get_queues())


class LotteryLottery(Lottery):
    """Run lottery on both non-empty and empty queues.

    Run lottery on all queues. When we get an non-empty queue, return
    immediately. If we get 10 empty queues in a row, run lottery again,
    and long poll on whatever queue we get.
    """

    def get_queue(self, *args, **kwargs):
        retry_empty_queue_count = 10

        for i in range(retry_empty_queue_count):
            queue = self._run_lottery(self.queue_info.get_queues())
            if self.queue_info.does_queue_need_work(queue):
                return queue
        return self._run_lottery(self.queue_info.get_queues())


class ReducedLottery(Lottery):
    """Improved lottery scheduling.

    Limiting the lottery pool by removing known empty queues. When we get an
    non-empty queue, return immediately. If we get an empty queue, we'll remove
    this empty queue out of the lottery pool and rerun lottery again. If all
    queues are empty, run lottery on all queues, and long poll on whatever
    queue we get.
    """

    def get_queue(self, *args, **kwargs):
        # Make a new copy of list, so no side effect on queue_info.queues
        candidate_queues = self.queue_info.get_queues()[:]

        while len(candidate_queues) > 0:
            queue = self._run_lottery(candidate_queues)
            if self.queue_info.does_queue_need_work(queue):
                return queue
            else:
                candidate_queues.remove(queue)
        return self._run_lottery(self.queue_info.get_queues())



"""Simple package for encrypting/decrypting strings.

This package provides a simple mechanism for encrypting and decrypting strings
very quickly using a private key. For simplicity, this key is stored in the
settings module and is used globally here -- that is, every message encrypted
with these functions will be encrypted with the same key.

This package is very simple. If you choose to encrypt data one day, and not
encrypt it the next day, you need to handle the failure scenarios.

Usage ::

    from kale import crypt

    encrypted_message = crypt.encrypt("foo")
    decrypted_message = crypt.decrypt(encrypted_message)

"""
from __future__ import absolute_import

import base64
import builtins
import codecs
import logging
import struct

from Crypto.Cipher import AES
import six

from kale import settings

log = logging.getLogger(__name__)

# Set up our Crypto hashing object
BLOCK_SIZE = 16


def _set_cipher():
    global cipher
    cipher = AES.new(builtins.bytes(settings.UTIL_CRYPT_CIPHER, 'ascii'), AES.MODE_ECB)


# Get our cipher key from the settings file
cipher = None
if settings.UTIL_CRYPT_CIPHER:
    _set_cipher()


class CryptException(Exception):
    """General encryption/decryption failure exception."""


def encrypt(msg):
    """Encrypts a message.

    :param msg: byte string message to be encrypted.
    :return: bytes for an encrypted version of msg.
    :raises: ValueError if passed anything other than bytes.
        CryptException if the encryption fails.
    """
    if not cipher:
        return msg
    if not isinstance(msg, six.binary_type):
        raise ValueError('only bytes can be encrypted')
    if six.PY2:
        msg = builtins.bytes(msg)

    msg = _pad(msg)
    msg = cipher.encrypt(msg)
    msg = base64.b64encode(msg)

    return msg


def decrypt(msg):
    """Decrypts a message.

    :param bytes msg: string of message to be decrypted.
    :return: bytes for original message.
    :rtype: bytes
    """
    if not cipher:
        return msg

    if isinstance(msg, six.text_type):
        # This should be a base64 string, so it
        # should encode to ascii without any problems.
        msg = msg.encode('ascii')

    if not isinstance(msg, six.binary_type):
        raise ValueError('Only bytes(or unicodes) can be decrypted')

    if six.PY2:
        msg = builtins.bytes(msg)

    try:
        msg = _unpad(cipher.decrypt(base64.b64decode(msg)))
    except (ValueError, TypeError) as e:
        # We can get struct.error if we end up passing an empty string
        # to _unpad. We get
        raise CryptException(e)

    return msg


def urlsafe_encrypt(msg):
    """Urlsafe encrypts a message.

    :param bytes msg: string message to be encrypted.
    :return: string of encrypted version of msg.
    :rtype: bytes
    """
    if not cipher:
        return msg
    if not isinstance(msg, six.binary_type):
        raise ValueError('only bytes can be encrypted')

    msg = base64.urlsafe_b64encode(cipher.encrypt(_pad(msg)))

    return msg


def hex_encrypt(msg):
    """Hex encrypts a message.

    :param bytes msg: string message to be encrypted.
    :return: string for encrypted version of msg in hex.
    :rtype: bytes
    """
    if not cipher:
        return msg
    if not isinstance(msg, six.binary_type):
        raise ValueError('only bytes can be encrypted')

    msg = cipher.encrypt(_pad(msg))
    msg = codecs.encode(msg, 'hex')

    return msg


def hex_decrypt(msg):
    """Decrypts a message.

    :param bytes msg: string for the message to be decrypted.
    :return: string for the original message.
    :rtype: bytes
    """
    if not cipher:
        return msg

    if isinstance(msg, six.text_type):
        # This should be a hex encoded string, so it
        # should encode to ascii without any problems.
        msg = msg.encode('ascii')

    if not isinstance(msg, six.binary_type):
        raise ValueError('Only bytes or unicodes can be decrypted')

    try:
        msg = _unpad(cipher.decrypt(codecs.decode(msg, 'hex')))
    except (ValueError, TypeError) as e:
        raise CryptException(e)

    return msg


def urlsafe_decrypt(msg):
    """Urlsafe decrypts a message.

    :param bytes msg: string for the message to be decrypted.
    :return: string for the original message.
    :rtype: bytes
    """

    if not cipher:
        return msg

    if isinstance(msg, six.text_type):
        # This should be a base64 encoded string, so it
        # should encode to ascii without any problems.
        msg = msg.encode('ascii')

    if not isinstance(msg, six.binary_type):
        raise ValueError('Only bytes(or unicodes) can be decrypted')

    try:
        msg = _unpad(cipher.decrypt(base64.urlsafe_b64decode(msg)))
    except (ValueError, TypeError) as e:
        raise CryptException(e)

    return msg


def _pad(msg):
    """Pad the message with enough bytes to be a multiple of BLOCK_SIZE.

    :param bytes msg: bytes message to be padded.
    :return: the msg with padding added.
    :rtype: bytes
    """

    padding_bytes = _get_padding_bytes(len(msg))
    return msg + padding_bytes


def _get_padding_bytes(msg_length):
    """Gets the bytes to pad a message of a given length with.

    We pad out with null bytes until we are one short of being
    a multiple of BLOCK_SIZE. The last byte is always the number
    of bytes we padded out(including the last one). We always apply
    some padding in order to be able to consistently remove the padding,
    so if we would apply no padding, we add padding of
    BLOCK_SIZE.

    :param msg_length: the length of the message to pad.
    :return: the bytes that should be appended to the message to pad it.
    """

    bytes_to_pad = BLOCK_SIZE - msg_length % BLOCK_SIZE
    if bytes_to_pad == 0:
        bytes_to_pad = BLOCK_SIZE
    pack_format = '%sB' % ('x' * (bytes_to_pad - 1))
    msg = struct.pack(pack_format, bytes_to_pad)
    return builtins.bytes(msg)


def _unpad(msg):
    """Removes the padding we applied to the message with _pad.

    We read the last byte of the message as an integer
    and strip that many bytes off the end of the message. Note that
    reading as a regular integer(i.e. struct.unpack('i')) would
    read multiple bytes off the end which would be bad if,
    for instance, we only had one byte of padding.

    :param msg: bytes or unicode msg that needs to have padding stripped from it.
    :return: bytes for the message with the padding removed.
    """
    if not msg or len(msg) < BLOCK_SIZE:
        raise ValueError('decrypted message was not padded correctly')
    msg = builtins.bytes(msg)
    bytes_to_strip = int(msg[-1])
    # If we are trying to strip off more than BLOCK_SIZE bytes,
    # or more bytes than there are in the msg,
    # something has gone wrong, likely the msg was corrupt.
    if bytes_to_strip > BLOCK_SIZE or bytes_to_strip > len(msg):
        raise ValueError('decrypted message was not padded correctly')
    return msg[:-bytes_to_strip]



"""Exceptions raised by the kale package."""


class ImproperlyConfiguredException(Exception):
    """Exception raised when Kale is improperly configured."""


class TaskException(Exception):
    """Base class for task exceptions."""


class ChubbyTaskException(TaskException):
    """Exception raised when a task is too chubby for SQS."""


class TimeoutException(TaskException):
    """Exception raised when a task exceeds its time limit."""
    pass


class InvalidTimeLimitTaskException(TaskException):
    """Exception raised when a task's time_limit exceeds its queue's
       visibility timeout.
    """


class InvalidTaskDelayException(TaskException):
    """Exception raised when a task is published with an invalid delay time."""


class BlacklistedException(TaskException):
    """Exception raised when a task has been blacklisted."""



"""Module containing utility functions for kale."""
import resource
import sys


def class_import_from_path(path_to_class):
    """Import a class from a path string.

    :param str path_to_class: class path, e.g., kale.consumer.Consumer
    :return: class object
    :rtype: class
    """

    components = path_to_class.split('.')
    module = __import__('.'.join(components[:-1]))
    for comp in components[1:-1]:
        module = getattr(module, comp)
    return getattr(module, components[-1])


def ru_maxrss_mb():
    """Gets memory residence set size in megabytes.

    Returns:
        Integer, megabytes.
    """
    resource_data = resource.getrusage(resource.RUSAGE_SELF)
    if sys.platform == 'darwin':
        return resource_data.ru_maxrss / (1024 * 1024)
    else:
        return resource_data.ru_maxrss / 1024



"""Module containing task publishing functionality."""
from __future__ import absolute_import

import logging

from kale import exceptions
from kale import message
from kale import settings
from kale import sqs
from kale import queue_info
from kale import utils

logger = logging.getLogger(__name__)


class Publisher(sqs.SQSTalk):
    """Class to manage publishing SQS tasks."""

    def publish(self, task_class, task_id, payload,
                current_retry_num=None, delay_sec=None):
        """Publish the given task type to the queue with the provided payload.

        :param obj task_class: class of the task that we are publishing.
        :param str task_id: unique identifying string for this task.
        :param dict payload: dictionary for the task payload.
        :param int current_retry_num: current task retry count. If 0, this is
            the first attempt to run the task.
        :param int delay_sec: time (in seconds) that a task should stay
                in the queue before being released to consumers.
        :raises: TaskTooChubbyException: This task is outrageously chubby.
                The publisher of the task should handle this exception and
                determine how to proceed.
        """

        if delay_sec is not None and delay_sec > settings.SQS_MAX_TASK_DELAY_SEC:
            raise exceptions.InvalidTaskDelayException(
                'Invalid task delay_sec (%d > %d).' % (
                    delay_sec, settings.SQS_MAX_TASK_DELAY_SEC))

        queue_class = utils.class_import_from_path(settings.QUEUE_CLASS)
        q_info = queue_info.QueueInfo(settings.QUEUE_CONFIG, self, queue_class)
        queue_obj = q_info.get_queue(task_class.queue)

        if task_class.time_limit >= queue_obj.visibility_timeout_sec:
            raise exceptions.InvalidTimeLimitTaskException(
                'Invalid task time limit: %d >= %d from %s' % (
                    task_class.time_limit, queue_obj.visibility_timeout_sec,
                    settings.QUEUE_CONFIG))

        queue = self._get_or_create_queue(queue_obj.name)
        sqs_msg = message.KaleMessage.create_message(
            task_class=task_class,
            task_id=task_id,
            payload=payload,
            queue=queue,
            current_retry_num=current_retry_num)

        queue.write(sqs_msg, delay_seconds=delay_sec)

        logger.debug('Published task. Task id: %s; Task name: %s' % (
            task_id, '%s.%s' % (task_class.__module__, task_class.__name__)))

    def publish_messages_to_dead_letter_queue(self, dlq_name, messages):
        """Sends a batch of messages to the dead letter queue.

        :param str dlq_name: dead-letter-queue name to send these messages to.
        :param list[KaleMessage] messages: a list of KaleMessage instances that
            have permanently failed.
        """
        dead_letter_queue = self._get_or_create_queue(dlq_name)

        # Message batches are a list of tuples where each tuple contains:
        #  1) A string that is unique to this list.
        #  2) The enconded message body.
        #  3) The task delay (always zero for the dead-letter-queue).
        message_batch = [(message.id, message.get_body_encoded(),
                          0) for message in messages]
        dead_letter_queue.write_batch(message_batch)



"""Module containing the base class for tasks."""
from __future__ import absolute_import

import logging
import time
import uuid

from kale import exceptions
from kale import publisher
from kale import settings

logger = logging.getLogger(__name__)

PERMANENT_FAILURE_UNRECOVERABLE = 'unrecoverable'
PERMANENT_FAILURE_NO_RETRY = 'no_retries'
PERMANENT_FAILURE_RETRIES_EXCEEDED = 'max_retries'


class Task(object):
    """Base class for kale tasks."""

    # Exponential retry settings. Overridable on a per-task basis based on
    # the task type, but these are our defaults.
    # Each time a task is retried its given a timeout before it will be
    # visible to consumers.
    # This is calculated with the following:
    # timeout = (1 << current_retries) * retry_delay_multiple
    _retry_delay_multiple = settings.RETRY_DELAY_MULTIPLE_SEC

    # Number of times a task can be retried after it fails the first time.
    # Set to None or 0 for no retries.
    max_retries = 4

    # Alert on tasks that exceed this time.
    target_runtime = 50

    # Default time until task is killed.
    time_limit = 60

    # Blacklist of exceptions to never
    # retry on (unlikely to be transient).
    EXCEPTION_LIST = (
        exceptions.BlacklistedException,
        KeyError,
        NotImplementedError,
        SyntaxError,
        UnboundLocalError)

    # String representation the a task queue, each task must override
    # this value.
    queue = None

    def __init__(self, message_body=None, *args, **kwargs):
        """Initialize an instance of a task.

        :param message_payload: Payload this task is being created from
                (optionally None for testability).
        """
        # global mocking
        message_body = message_body or {}
        self.task_id = message_body.get('id')
        self.task_name = '%s.%s' % (self.__class__.__module__,
                                    self.__class__.__name__)
        self.app_data = None
        payload = message_body.get('payload')
        if payload:
            self.app_data = payload.get('app_data', None)

        # Used for tracking/diagnostics.
        self._publisher_data = message_body.get('_publisher')
        self._enqueued_time = message_body.get('_enqueued_time', 0)
        self._dequeued_time = None
        self._start_time = None
        self._end_time = None
        self._task_latency_sec = 0

        self._dequeued()

    @classmethod
    def _get_task_id(cls, *args, **kwargs):
        """Return a unique task identifier.

        This can be overridden in sub-classes to create task ids that
        are more descriptive.

        :return: the unique identifier for a task.
        :rtype: str
        """
        return '%s_uuid_%s' % (cls.__name__, uuid.uuid1())

    @classmethod
    def publish(cls, app_data, *args, **kwargs):
        """Class method to publish a task given instance specific arguments."""
        delay_sec = kwargs.get('delay_sec')
        task_id = cls._get_task_id(*args, **kwargs)
        payload = {
            'args': args,
            'kwargs': kwargs,
            'app_data': app_data}
        pub = publisher.Publisher()
        pub.publish(cls, task_id, payload, delay_sec=delay_sec)
        return task_id

    @classmethod
    def _get_delay_sec_for_retry(cls, current_retry_num):
        """Generate a delay based on the number of times the task has failed.

        :param int current_retry_num: Task retry count for the task that is
            about to be published.
        :return: Number of seconds this task should wait before running again.
        :rtype: int
        """
        # Exponentially backoff the wait time for task attempts.
        return min(((1 << current_retry_num) * cls._retry_delay_multiple),
                   settings.SQS_MAX_TASK_DELAY_SEC)

    @classmethod
    def handle_failure(cls, message, raised_exception):
        """Logic to respond to task failure.

        :param KaleMessage message: instance of KaleMessage containing the
            task that failed.
        :param Exception raised_exception: exception that the failed task
            raised.
        :return: True if the task will be retried, False otherwise.
        :rtype: boolean
        """

        logger.warning('Task %s failed: %s.' % (message.task_id,
                                                raised_exception))

        if isinstance(raised_exception, exceptions.TimeoutException):
            message.task_inst._kill_runtime_exceeded()

        # If our exception falls into a specific list, we bail out completely
        # and do not retry.
        if isinstance(raised_exception, cls.EXCEPTION_LIST):
            cls._report_permanent_failure(
                message, raised_exception,
                PERMANENT_FAILURE_UNRECOVERABLE, True)
            return False

        # See if retries are enabled at all. If is <= 0, then just return.
        if cls.max_retries is None or cls.max_retries <= 0:
            cls._report_permanent_failure(
                message, raised_exception, PERMANENT_FAILURE_NO_RETRY, True)
            return False

        # Monitor retries and dropped tasks
        if message.task_retry_num >= cls.max_retries:
            cls._report_permanent_failure(
                message, raised_exception,
                PERMANENT_FAILURE_RETRIES_EXCEEDED, False)
            return False

        payload = {
            'args': message.task_args,
            'kwargs': message.task_kwargs,
            'app_data': message.task_app_data}
        retry_count = message.task_retry_num + 1
        delay_sec = cls._get_delay_sec_for_retry(message.task_retry_num)
        pub = publisher.Publisher()
        pub.publish(
            cls, message.task_id, payload,
            current_retry_num=retry_count, delay_sec=delay_sec)
        return True

    def run(self, *args, **kwargs):
        """Wrap the run_task method of tasks.

        We use this instead of a decorator to protect against the case where
        a subclass may not call super().

        The order of operations in a subclass should look like this:
        1) Subclass's override _pre_run logic.
        2) call super() at the end of the override.
        3) run_task() willrun.
        4) call super() at the start of _post_run override.
        5) Subclass's override _post_run logic.
        """

        self._setup_task_environment()
        self._pre_run(*args, **kwargs)

        try:
            # This raises an exception if the task should not be attempted.
            # This enables tasks to be blacklisted by ID or type.
            self._check_blacklist(*args, **kwargs)
            self.run_task(*args, **kwargs)
        except Exception as exc:
            # Record latency here.
            self._end_time = time.time()
            self._task_latency_sec = self._end_time - self._start_time

            # Cleanup the environment this task was running in right away.
            self._clean_task_environment(task_id=self.task_id,
                                         task_name=self.task_name, exc=exc)
            raise

        self._post_run(*args, **kwargs)
        self._clean_task_environment(task_id=self.task_id,
                                     task_name=self.task_name)

    def run_task(self, *args, **kwargs):
        """Run the task, this must be implemented by subclasses."""
        raise NotImplementedError()

    def _check_blacklist(self, *args, **kwargs):
        """Raises an exception if a task should not run.

        This enables a subclass to blacklist tasks by ID or type. This needs
        to be handled outside of kale since it requires a datastore.

        Ex:
        if task_id in cache.get(BLACKLISTED_IDS) or task_name in
            cache.get(BLACKLISTED_TASK_TYPES):
                raise exceptions.BlacklistedException()
        """
        return

    def _dequeued(self, *args, **kwargs):
        """Method called when a task is pulled from the queue and instantiated.

        This does not mean that this task instance  will necessarily run. It
        was most likely pulled from the queue in a batch and will still sit
        idle until its turn, if it's visibility timeout runs out before that
        point it will be released back to the queue.

        Note: We do funny things in celery to determine time spent enqueued,
        in SQS we may have alternative options here that require
        the message but no args/kwargs.
        """
        self._dequeued_time = time.time()

    @staticmethod
    def _setup_task_environment():
        """Setup the environment for this task."""
        pass

    @staticmethod
    def _clean_task_environment(task_id=None, task_name=None, exc=None):
        """Cleans the environment for this task.

        Args:
            task_id: string of task id.
            task_name: string of task name.
            exc: The exception raised by the task, None if the task succeeded.
        """
        pass

    def _pre_run(self, *args, **kwargs):
        """Called immediately prior to a task running."""
        self._start_time = time.time()

    def _post_run(self, *args, **kwargs):
        """Called immediately after a task finishes.

        This will not be called if the task fails.
        """
        self._end_time = time.time()
        self._task_latency_sec = self._end_time - self._start_time

        if self.target_runtime:
            if self._task_latency_sec >= self.target_runtime:
                self._alert_runtime_exceeded()

    def _alert_runtime_exceeded(self, *args, **kwargs):
        """Handle the case where a task exceeds its alert runtime.

        This will be called during task post_run.
        """
        pass

    def _kill_runtime_exceeded(self, *args, **kwargs):
        """Handle the case where a task is killed due to timing out.

        This will be called by the task's onfailure
        """
        pass

    @classmethod
    def _report_permanent_failure(cls, message, exception, failure_type,
                                  log_exception):
        """Handles reporting of a permanent failure.

        :param KaleMessage message: A kale.message instance that contains the
            task.
        :param Exception exception: The error info that contains the stacktrace
        :param str failure_type:  A string used for logging to denote the type
            of permanent failure.
        :param bool log_exception: A bool denoting if we should include the
                exception stacktrace.
        """
        message_str = ('PERMANENT_TASK_FAILURE: FAILURE_TYPE=%s TASK_TYPE=%s '
                       'TASK_ID=%s, TRACEBACK=%s' %
                       (failure_type, message.task_name, message.task_id,
                        exception))
        logger.error(message_str)



"""Module containing task consumption functionality."""
from __future__ import absolute_import

from kale import sqs


class Consumer(sqs.SQSTalk):
    """SQS utility class for consuming tasks."""

    def fetch_batch(self, queue_name, batch_size, visibility_timeout_sec,
                    long_poll_time_sec=None):
        """Fetches a batch of messages from a queue.

        :param str queue_name: queue name.
        :param int batch_size: batch size.
        :param int visibility_timeout_sec: visibility timeout in seconds.
        :param int long_poll_time_sec: long poll time in seconds.
        :returns: a list of KaleMessage objects, or None if not message found.
        :rtype: list[KaleMessage]
        """
        queue = self._get_or_create_queue(queue_name)
        extra_options = {}
        if long_poll_time_sec:
            extra_options['wait_time_seconds'] = long_poll_time_sec
        messages = queue.get_messages(
            num_messages=batch_size,
            visibility_timeout=visibility_timeout_sec,
            **extra_options)
        if messages:
            # Call get_body on all of the messages to decrypt and instantiate
            # each task, this allows us to track each task's dequeue time.
            [message.get_body() for message in messages]
            return messages
        else:
            return None

    def delete_messages(self, messages, queue_name):
        """Remove messages from the queue.

        :param list[KaleMessage] messages: messages to delete.
        :param str queue_name: queue name.
        """
        if not messages:
            return
        queue = self._get_or_create_queue(queue_name)
        self._connection.delete_message_batch(queue, messages)

    def release_messages(self, messages, queue_name):
        """Releases messages to SQS queues so other workers can pick them up.

        :param list[KaleMessage] messages: messages to release to SQS.
        :param str queue_name: queue name.
        """
        if not messages:
            return
        queue = self._get_or_create_queue(queue_name)
        message_tups = [(msg, 0) for msg in messages]
        self._connection.change_message_visibility_batch(queue, message_tups)



"""Default settings for kale."""
from __future__ import absolute_import

import os
import pickle
import platform
import time
import zlib

import six

# The default settings are inadequate for actual use
# due to the need for a queue. When the settings are
# overridden this will be set to True.
# Note: You are not supposed to overwrite this property.
PROPERLY_CONFIGURED = False

# Optional functions to be called when the worker is
# started and stopped.
ON_WORKER_STARTUP = lambda: None
ON_WORKER_SHUTDOWN = lambda: None

# Note: This makes it possible to move modules and remain
# backwards compatible.
TASK_MAPPER = {}

# The function to (de)compress a message string, which takes a string as input.
COMPRESSOR = zlib.compress
DECOMPRESSOR = zlib.decompress

# Set to True (in dev) when connecting to ElasticMQ instead of SQS
MESSAGE_QUEUE_USE_PROXY = False
# Set to a valid proxy port when connecting to ElasticMQ instead of SQS
MESSAGE_QUEUE_PROXY_PORT = 0
MESSAGE_QUEUE_PROXY_HOST = ''

# AWS credential for connecting to SQS
AWS_ACCESS_KEY_ID = ''
AWS_SECRET_ACCESS_KEY = ''
AWS_REGION = ''

# Use max size (in bytes) as of March 2014 as the default.
SQS_TASK_SIZE_LIMIT = 256000

RETRY_DELAY_MULTIPLE_SEC = 60
SQS_MAX_TASK_DELAY_SEC = 900
TIMESTAMP_FUNC = time.time
PUBLISHER_STR_FUNC = lambda: '%s[%d]' % (platform.node(), os.getpid())

RESET_TIMEOUT_THRESHOLD = 1

USE_DEAD_LETTER_QUEUE = True

# Path for queue config
QUEUE_CONFIG = 'queue_config.yaml'
QUEUE_CLASS = 'kale.queue_info.TaskQueue'
QUEUE_SELECTOR = 'kale.queue_selector.ReducedLottery'

# We will gracefully stop this process if memory usage
# exceeds this amount (in MB).
DIE_ON_RESIDENT_SET_SIZE_MB = 256

# CIPHER used by kale.crypt, must be 16-, 24-, or 36-byte string
UTIL_CRYPT_CIPHER = '1234567890123456'

ENABLE_DEAD_LETTER_QUEUE = True

# Manually specify pickle protocol used for writing Pickle files
# Py2/Py3 have different default protocols. Reproduce those defaults
# here for backwards compatibility.
# Note: If Python 2 & Python 3 need to co-exist, override PICKLE_PROTOCOL=2
if six.PY2:
    PICKLE_PROTOCOL = 0
else:
    PICKLE_PROTOCOL = pickle.DEFAULT_PROTOCOL



"""Settings and configuration for Kale.

This will create a settings module that overrides default
settings (from the default_settings module) and it will override
those settings with values found in the module specified by the
KALE_SETTINGS_MODULE environment variable.

Any machine that wants to use these tasks MUST have
KALE_SETTINGS_MODULE as an environment variable or this module
will raise an exception.
"""
from __future__ import absolute_import

import importlib
import logging
import os

from kale import default_settings

logger = logging.getLogger(__name__)

ENVIRONMENT_VARIABLE = 'KALE_SETTINGS_MODULE'

_settings_module = None


def setup_package():
    global _settings_module
    _settings_module = os.environ.get(ENVIRONMENT_VARIABLE)
    os.environ[ENVIRONMENT_VARIABLE] = 'kale.tests.test_settings'
    # Re-intitialize settings
    global settings
    settings.__init__()


def teardown_package():
    global _settings_module
    if _settings_module:
        os.environ[ENVIRONMENT_VARIABLE] = _settings_module
    # Re-intitialize settings
    global settings
    settings.__init__()


class Settings(object):
    """Singleton class to manage kale settings."""

    _instance = None

    def __new__(cls, *args, **kwargs):
        """Set this class up as a singleton."""

        if not cls._instance:
            cls._instance = super(Settings, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self, *args, **kwargs):
        """Instantiate the settings."""

        # update this dict from default settings (but only for ALL_CAPS
        # settings)
        for setting in dir(default_settings):
            if setting == setting.upper():
                setattr(self, setting, getattr(default_settings, setting))

        try:
            settings_module_path = os.environ[ENVIRONMENT_VARIABLE]
        except KeyError:
            # NOTE: This is arguably an EnvironmentError, but that causes
            # problems with Python's interactive help.
            logger.error(
                ('Settings cannot be imported, because environment '
                 'variable %s is undefined.') % ENVIRONMENT_VARIABLE)
            return

        try:
            settings_module = importlib.import_module(settings_module_path)
        except ImportError as e:
            error = ImportError(
                'Could not import settings "%s" (Is it on sys.path?): %s' %
                (settings_module_path, e))
            logger.error(error)
            return

        for setting in dir(settings_module):
            if setting == setting.upper():
                setting_value = getattr(settings_module, setting)
                setattr(self, setting, setting_value)

        # This setting lets the application know that settings
        # have been properly configured.
        self.PROPERLY_CONFIGURED = True


# Instantiate the settings globally.
settings = Settings()



"""Helpful tools for testing kale tasks."""
from __future__ import absolute_import

from kale import exceptions
from kale import message
from kale import queue_info
from kale import queue_selector
from kale import settings
from kale import task


class MockTask(task.Task):

    @classmethod
    def _get_task_id(self, *args, **kwargs):
        return 'mock_task'

    def run_task(self, *args, **kwargs):
        pass


class FailTask(task.Task):

    @classmethod
    def _get_task_id(self, *args, **kwargs):
        return 'fail_task'

    def run_task(self, *args, **kwargs):
        raise exceptions.TaskException('Task failed.')


class TimeoutTask(task.Task):

    @classmethod
    def _get_task_id(self, *args, **kwargs):
        return 'fail_task'

    def run_task(self, *args, **kwargs):
        raise exceptions.TimeoutException('Task failed.')


class SlowButNotTooSlowTask(task.Task):

    time_limit = 100
    target_runtime = 90

    @classmethod
    def _get_task_id(self, *args, **kwargs):
        return 'moderately_slow_task'

    def run_task(self, *args, **kwargs):
        # Ugly trick.
        self._start_time = self._start_time - self.target_runtime


class FailTaskNoRetries(FailTask):

    max_retries = None


class MockMessage(message.KaleMessage):

    def __init__(self, task_inst, task_args=None, task_kwargs=None, app_data=None,
                 retry_num=0):
        """Instantiate a mock KaleMessage.

        Args:
            task: An instance of a task.
        """
        self.id = 'id'
        self.task_name = task_inst.task_name
        self.task_id = task_inst.task_id
        self.task_args = task_args or []
        self.task_kwargs = task_kwargs or {}
        self.task_app_data = app_data or {}
        self.task_retry_num = retry_num
        self.task_inst = task_inst


class MockConsumer(object):

    def consume(*args, **kwargs):
        return []


class TestQueueClass(queue_info.TaskQueue):
    pass


class TestQueueSelector(queue_selector.SelectQueueBase):

    def __init__(self, queue_info):
        self.queue_info = queue_info

    def get_queue(self, *args, **kwargs):
        """Returns a TaskQueue object."""
        return self.queue_info.get_queues()[0]


def new_mock_message(task_class=None):
    """Create a new mock message.

    Args:
        task_class: Task class to use in message (default is MockTask).
    """
    task_inst = new_mock_task(task_class)
    message = MockMessage(task_inst)
    return message


def new_mock_task(task_class=None):
    """Create a new mock task instance.

    Args:
        task_class: Task class to use in message (default is MockTask).
    """
    task_class = task_class if task_class else MockTask

    mock_payload = {
        'id': task_class._get_task_id(),
        '_enqueued_time': settings.TIMESTAMP_FUNC(),
        '_publisher': settings.PUBLISHER_STR_FUNC()}
    return task_class(mock_payload)



"""Custom message type for SQS messages."""
from __future__ import absolute_import

import pickle

from boto.sqs import message
import six

from kale import crypt
from kale import exceptions
from kale import settings
from kale import utils

_compressor = settings.COMPRESSOR
_decompressor = settings.DECOMPRESSOR
_task_size_limit = settings.SQS_TASK_SIZE_LIMIT
_get_current_timestamp = settings.TIMESTAMP_FUNC
_get_publisher_data = settings.PUBLISHER_STR_FUNC


class KaleMessage(message.RawMessage):
    """Kale message representing the data stored in an SQS queue."""

    # _task_mapper is a class cache mapping task paths to classes.
    # It will initially be populated with keys provided and will lazily
    # create other mappings.
    _task_mapper = None

    def __init__(self, queue=None, body=''):
        """Constructor.

        :param queue:
        :param body: string for message body
        """
        message.RawMessage.__init__(self, queue=queue, body=body)

        # Lazily instantiate the task mapper.
        if not self._task_mapper:
            self._task_mapper = {k: utils.class_import_from_path(v)
                                 for k, v in six.iteritems(settings.TASK_MAPPER)}

    @classmethod
    def create_message(cls, task_class=None, task_id=None, payload=None,
                       queue=None, current_retry_num=None):
        """Factory method to craete a message.

        :param task_class: class of a task.
        :param task_id: string for task id.
        :param payload: dictionary for message payload.
        :param queue: object of boto.sqs.queue.Queue.
        :param current_retry_num: integer for how many retries so far.
        :return:
        """
        # Consider moving the queue to be a task_class
        # attr (needs to be fetched from SQS though).
        KaleMessage._validate_task_payload(payload)
        retry_count = current_retry_num or 0
        message_data = {
            'id': task_id,
            # This represents the path to the task. The consumer will have a
            # dictionary mapping these values to task classes.
            'task': '%s.%s' % (task_class.__module__, task_class.__name__),
            # Payload holds the data that the task's run_task method will be
            # called with.
            # Ex: mytask.ThisTask().run_task(
            # *payload['args'], **payload['kwargs'])
            'payload': payload,
            # These are task attributes used for tracking/analytics/debugging
            # Time when the task publisher published this task.
            '_enqueued_time': _get_current_timestamp(),
            # Publishing host (potentially include caller function)
            '_publisher': _get_publisher_data(),
            # Current task retry. This will be 0 from new tasks and will be
            # incremented for each retry.
            'retry_num': retry_count,
        }
        return KaleMessage(queue=queue, body=message_data)

    @staticmethod
    def _validate_task_payload(payload):
        """Validate that this is a valid task.

        :param payload: dictionary that will be submitted to the queue.
        :raises: AssertionError if payload is invalid.
        """

        assert 'args' in payload, 'args is required to be in the payload'
        assert 'kwargs' in payload, 'kwargs is required to be in the payload'

    def encode(self, msg):
        """Custom encoding for Kale tasks.

        :param dict msg: message to decode.
        :return: string for encoded message.
        :rtype: str
        """

        compressed_msg = _compressor(pickle.dumps(msg, protocol=settings.PICKLE_PROTOCOL))
        compressed_msg = crypt.encrypt(compressed_msg)
        # Check compressed task size.
        if len(compressed_msg) >= _task_size_limit:
            task_id = msg.get('task_id')
            raise exceptions.ChubbyTaskException(
                'Task %s is over the limit of %d bytes.' % (task_id,
                                                            _task_size_limit))

        return compressed_msg

    def decode(self, msg):
        """Custom decoding for Kale tasks.

        :param str msg: message to decode.
        :return: dictionary for decoded message.
        :rtype: dict
        """

        message_body = crypt.decrypt(msg)
        message_body = pickle.loads(_decompressor(message_body))
        self.task_name = message_body['task']
        self.task_id = message_body['id']
        self.task_args = message_body['payload']['args']
        self.task_kwargs = message_body['payload']['kwargs']
        self.task_retry_num = message_body['retry_num']
        self.task_app_data = message_body['payload'].get('app_data')

        # This will instantiate the task.
        self.task_inst = self._class_from_path(self.task_name)(message_body)

        return message_body

    def _class_from_path(self, task_path):
        """Return the task class given a task's path.

        :param task_path: string for a class, e.g., mytask.MyTask
        :return:
        """
        if task_path not in self._task_mapper:
            task_class = utils.class_import_from_path(task_path)
            self._task_mapper[task_path] = task_class
        return self._task_mapper[task_path]



"""Base class for SQS utility classes."""
from __future__ import absolute_import

import logging

import boto.sqs
import boto.sqs.connection
import boto.sqs.regioninfo

from kale import exceptions
from kale import message
from kale import settings

logger = logging.getLogger(__name__)


class SQSTalk(object):
    """Base class for SQS utility classes."""

    # Class attribute for storing connections.
    _connections = {}
    _queues = {}

    def __init__(self, *args, **kwargs):
        """Constructor.
        :raises: exceptions.ImproperlyConfiguredException: Raised if the
            settings are not adequately configured.
        """

        if not settings.PROPERLY_CONFIGURED:
            raise exceptions.ImproperlyConfiguredException(
                'Settings are not properly configured.')

        aws_region = settings.AWS_REGION
        aws_access_key_id = settings.AWS_ACCESS_KEY_ID
        aws_secret_access_key = settings.AWS_SECRET_ACCESS_KEY

        conn_str = '%s:%s:%s' % (aws_region, aws_access_key_id,
                                 aws_secret_access_key)

        if conn_str not in self._connections:
            if settings.MESSAGE_QUEUE_USE_PROXY:
                # Used only in Dev environment which uses ElasticMQ instead of
                # SQS. Hence it uses a different boto api to connect to a proxy
                region = boto.sqs.regioninfo.RegionInfo(
                    name='proxy',
                    endpoint=settings.MESSAGE_QUEUE_PROXY_HOST)
                self._connections[conn_str] = boto.connect_sqs(
                    aws_access_key_id=aws_access_key_id,
                    aws_secret_access_key=aws_secret_access_key,
                    region=region,
                    is_secure=False,
                    port=settings.MESSAGE_QUEUE_PROXY_PORT)
            else:
                # Used in staging and production
                self._connections[conn_str] = boto.sqs.connect_to_region(
                    aws_region,
                    aws_access_key_id=aws_access_key_id,
                    aws_secret_access_key=aws_secret_access_key)

        self._connection = self._connections[conn_str]

    def _get_or_create_queue(self, queue_name):
        """Fetch or create a queue.

        :param str queue_name: string for queue name.
        :return: object of boto.sqs.queue.Queue.
        :rtype: Queue
        """
        # Check local cache first.
        if queue_name in self._queues:
            return self._queues[queue_name]

        queue = self._connection.lookup(queue_name)
        # If queue doesn't exist, create it.
        if queue is None:
            logger.info('Creating new SQS queue: %s' % queue_name)
            queue = self._connection.create_queue(queue_name)
        # Set the message class to be RawMessage so that
        # messages aren't decoded when fetched.
        queue.set_message_class(message.KaleMessage)
        self._queues[queue_name] = queue
        return queue

    def get_all_queues(self, prefix=''):
        """Returns all queues, filtered by prefix.

        :param str prefix: string for queue prefix.
        :return: a list of queue objects.
        :rtype: list[Queue]
        """
        return self._connection.get_all_queues(prefix)



"""Module containing timeout methods for monitoring tasks.

Modeled after:
http://stackoverflow.com/a/601168/854976
"""
from __future__ import absolute_import

import errno
import os
import signal
from contextlib import contextmanager

from kale import exceptions


@contextmanager
def time_limit(seconds, error_message=os.strerror(errno.ETIME)):
    """Context manager for handling method timeouts.

    Usage:
        try:
            with time_limit(10):
                some_function()
        except TimeoutException:
            # error handling here

    :param int seconds: seconds before timeout.
    :param str error_message: error message.
    :raises: kale.exceptions.TimeoutException
    """

    def _handle_timeout(signum, frame):
        """Handle timout signal."""
        raise exceptions.TimeoutException(error_message)

    original_handler = signal.signal(signal.SIGALRM, _handle_timeout)
    signal.alarm(seconds)
    try:
        yield
    finally:
        # Reset original handler.
        signal.signal(signal.SIGALRM, original_handler)
        # Remove signal.
        signal.alarm(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import multiprocessing  # To make python setup.py test happy
import os
import shutil
import subprocess
from distutils.command.clean import clean

from setuptools import setup

multiprocessing

PACKAGE = 'wrk'
__version__ = None
with open(os.path.join('kale', 'version.py')) as f:
    source = f.read()
code = compile(source, os.path.join('kale', 'version.py'), 'exec')
exec(code)  # set __version__


# -*- Hooks -*-

class CleanHook(clean):

    def run(self):
        clean.run(self)

        def maybe_rm(path):
            if os.path.exists(path):
                shutil.rmtree(path)

        if self.all:
            maybe_rm('wrk.egg-info')
            maybe_rm('build')
            maybe_rm('dist')
            subprocess.call('rm -rf *.egg', shell=True)
            subprocess.call('find . -name "*.pyc" -exec rm -rf {} \;',
                            shell=True)


# -*- Classifiers -*-

classes = """
    Development Status :: 5 - Production/Stable
    License :: OSI Approved :: BSD License
    Topic :: System :: Distributed Computing
    Topic :: Software Development :: Object Brokering
    Programming Language :: Python
    Programming Language :: Python
    Programming Language :: Python :: 2.7
    Programming Language :: Python :: 3.5
    Programming Language :: Python :: 3.6
    Programming Language :: Python :: 3.7
    Programming Language :: Python :: Implementation :: CPython
    Operating System :: OS Independent
"""
classifiers = [s.strip() for s in classes.split('\n') if s]

# -*- %%% -*-

setup(
    name=PACKAGE,
    version=__version__,
    description='Kale: A Task Worker Library from Nextdoor',
    long_description_content_type=open('README.md').read(),
    author='Nextdoor',
    author_email='eng@nextdoor.com',
    url='https://github.com/Nextdoor/wrk',
    download_url='http://pypi.python.org/pypi/wrk#downloads',
    license='Apache License, Version 2',
    keywords='kale nextdoor taskworker sqs python',
    packages=['kale'],
    tests_require=[
        'mock',
        'nose'
    ],
    test_suite='nose.collector',
    install_requires=[
        'boto',
        'pycryptodome',
        'pyyaml',
        'setuptools',
        'six',
        'future',
    ],
    classifiers=classifiers,
    cmdclass={'clean': CleanHook},
)



"""Get information of queues.

We can have different ways of getting queue information, e.g., from Zookeeper,
 from hardcoded data, from config file, ...

Thus, we need to extend the base class QueueInfoBase for different
implementations and use the right implementation in different cases.
"""
from __future__ import absolute_import

import six
import yaml
from boto import exception as boto_exception


class TaskQueue(object):
    """Represents a task queue. Always created via QueueInfo.


    :param str name: string for queue name.
    :param int priority: integer for queue priority.
    :param int num_iterations: integer for number of iterations to process
        tasks for a select_queue() call. Number of iterations to process
        tasks for a select_queue() call. Two reasons:
            1. Calling select_queue() has overhead, so we want to reuse
               the chosen queue for several rounds of tasks processing.
            2. SQS has limit of 10 tasks per batch. Using multiple
               iterations here simulates a larger batch, e.g., 20.
    :param int long_poll_time_sec: integer for number of seconds waiting for
        long poll. For SQS, if a queue is empty when task worker wants to
        fetch tasks from it, the task worker can wait for quite a while,
        hoping tasks would appear in the queue. This is to avoid costly
        connection reestablishment.
    :param int batch_size: integer for number of tasks to fetch in a batch.
    :param int visibility_timeout_sec: integer for the max time for a task that
            are invisible in an SQS queue. How to decide this property?
            - T: Average running time of a task.
            - N: Number of tasks in a batch that are fetched at once by
                 the worker.
            - F: Just to give us more headroom. E.g., 1.2
            => visibility_timeout = F * (T * N)
    """

    def __init__(self, name='default', priority=5, num_iterations=2,
                 long_poll_time_sec=5, batch_size=5,
                 visibility_timeout_sec=600):
        self.simple_name = name
        (self.name, self.dlq_name) = self._decorate_name(name)
        self.priority = priority
        self.num_iterations = num_iterations
        self.batch_size = batch_size
        self.long_poll_time_sec = long_poll_time_sec
        self.visibility_timeout_sec = visibility_timeout_sec

    def _decorate_name(self, name):
        """Decorate queue name.

        For example, we want to prefix RELEASE version on each queue name.
        The default implementation is to do nothing.

        :param str name: string for queue name.
        :return: A 2-tuple (queue name, dead letter queue name).
        :rtype: tuple
        """
        return (name, 'dlq-' + name)


class QueueInfoBase(object):
    """Base class to represent Queue information.

    Any concrete class should implement get_queues method.
    """

    def get_queues(self):
        """Returns a list of TaskQueue objects."""
        raise NotImplementedError('Base class cannot be used directly.')

    def get_highest_priority_queue_that_needs_work(self):
        """Returns the highest-priority non-empty queue."""
        raise NotImplementedError('Base class cannot be used directly.')

    def is_queue_empty(self, queue):
        """Check if a queue is empty.

        :param TaskQueue queue: A TaskQueue object.
        :return: True if the queue is empty; otherwise, queue is non-empty.
        :rtype: bool
        """
        raise NotImplementedError('Base class cannot be used directly.')

    def does_queue_need_work(self, queue):
        """Checks if a queue should be worked on.

        :param TaskQueue queue: a TaskQueue object.
        :return: True if the queue needs work; False otherwise.
        :rtype: bool
        """
        raise NotImplementedError('Base class cannot be used directly.')


class QueueInfo(QueueInfoBase):
    """Provides task queue info."""

    _queues = None
    _simple_name_queues_map = None

    def __init__(self, config_file, sqs_talk, queue_cls=TaskQueue):
        """Instantiate new QueueInfo object.

        :param str config_file: String of config file path.
        :param SQSTalk sqs_talk: An SQSTalk object.
        :param TaskQueue queue_cls: Class (or subclass) of TaskQueue.
        """

        # Initialize singleton if needed
        if not QueueInfo._queues:
            QueueInfo._queues = self._get_queues_from_config(config_file,
                                                             queue_cls)

        if not QueueInfo._simple_name_queues_map:
            QueueInfo._simple_name_queues_map = {}
            for queue in QueueInfo._queues:
                QueueInfo._simple_name_queues_map[queue.simple_name] = queue

        self._sqs_talk = sqs_talk

    def get_queue(self, simple_name):
        """Get queue object by simple name.

        :param str simple_name: string of queue simple name.
        :return: a TaskQueue object.
        :rtype: TaskQueue
        """
        return self._simple_name_queues_map[simple_name]

    def get_queues(self):
        """Returns a list of TaskQueue objs sorted by priority (highest first).
        """
        return self._queues

    def get_highest_priority_queue_that_needs_work(self):
        """Returns the highest-priority queue that needs work.

        If all queues are empty, then None is returned.
        """
        for queue in self._queues:
            if self.does_queue_need_work(queue):
                return queue
        return None

    def is_queue_empty(self, queue):
        """Check if a queue is empty.

        :param TaskQueue queue: a TaskQueue object.
        :return: True if the queue is empty; otherwise, queue is non-empty.
        :rtype: bool
        """
        sqs_queue = self._sqs_talk._get_or_create_queue(queue.name)
        if sqs_queue.count() > 0:
            return False
        return True

    def does_queue_need_work(self, queue):
        """Checks if a queue should be worked on.

        This basically checks whether the queue is empty. However,
        if we hit a SQS rate limit, this will assume the queue needs work.

        :param TaskQueue queue: a TaskQueue object.
        :return: True if the queue needs work; False otherwise.
        :rtype: bool
        """
        try:
            return not self.is_queue_empty(queue)
        except boto_exception.SQSError as e:
            if e.code == 'RequestThrottled':
                return True
            raise e

    @classmethod
    def _get_queues_from_config(cls, config_file, queue_cls):
        """Parses config file and returns queues.

        :param str config_file: String for the path of yaml config file for
            queues.
        :param TaskQueue queue_cls: Class (or subclass) of TaskQueue.
        :return: A list of TaskQueue queue objects sorted by priority.
            Highest priority first.
        :rtype: list[TaskQueue]
        """
        with open(config_file, 'r') as fp:
            queues_from_config = yaml.load(fp, Loader=yaml.FullLoader)
            queues = []
            for queue_name, queue in six.iteritems(queues_from_config):
                q = queue_cls(name=queue['name'], priority=queue['priority'],
                              num_iterations=queue['num_iterations'],
                              long_poll_time_sec=queue['long_poll_time_sec'],
                              batch_size=queue['batch_size'],
                              visibility_timeout_sec=queue[
                                  'visibility_timeout_sec'])
                queues.append(q)
            # Sort by priority, highest priority first.
            queues.sort(key=lambda x: x.priority, reverse=True)
            return queues



"""Module for running the worker process.

It's an infinite loop.
"""
from __future__ import absolute_import

import logging
import signal
import sys
import time

from six.moves import range

from kale import consumer
from kale import publisher
from kale import queue_info
from kale import settings
from kale import timeout
from kale import utils

logger = logging.getLogger(__name__)

SIGNALS_TO_HANDLE = (
    signal.SIGABRT,
    # This will catch Ctrl-C interrupts.
    signal.SIGINT,
    signal.SIGQUIT,
    signal.SIGTERM,
    # Handle Ctrl-Z/suspend.
    signal.SIGTSTP)

# Logging constants.
LOG_TASK_RESULT_DEFERRED = 'deferred'
LOG_TASK_RESULT_ERROR = 'error'
LOG_TASK_RESULT_SUCCESS = 'success'


class Worker(object):

    def __init__(self):
        """Initialize a worker instance."""

        self._consumer = consumer.Consumer()

        queue_class = utils.class_import_from_path(settings.QUEUE_CLASS)
        q_info = queue_info.QueueInfo(
            config_file=settings.QUEUE_CONFIG,
            sqs_talk=self._consumer,
            queue_cls=queue_class)

        queue_selector_class = utils.class_import_from_path(
            settings.QUEUE_SELECTOR)
        self._queue_selector = queue_selector_class(q_info)

        # The worker will publish permanently failed tasks to a
        # dead-letter-queue.
        self._publisher = publisher.Publisher()

        # Track total messages processed.
        self._total_messages_processed = 0

        # Intialize queue variables used by each batch.
        self._incomplete_messages = []
        self._successful_messages = []
        self._failed_messages = []
        self._permanent_failures = []
        self._batch_stop_time = time.time()
        self._batch_queue = None
        # Monitors whether the worker has been exposed to tasks and may
        # have bloated in memory.
        self._dirty = False

        # Setup signal handling for cleanup.
        for sig in SIGNALS_TO_HANDLE:
            signal.signal(sig, self._cleanup_worker)

        # Allow the client of this library to do any setup before
        # starting the worker.
        settings.ON_WORKER_STARTUP()

    def _on_pre_run_worker(self):
        """Callback function right at the beginning of starting the worker. """
        logger.info('Starting run loop for task worker.')

    def _on_exceeding_memory_limit(self, ru_maxrss_mb):
        """Callback function when the process exceeds memory limit.

        :param int ru_maxrss_mb: maximum resident set size used (in MB).
        """
        logger.info('Memory usage of %d MB exceeds max of %s MB. Exiting.' % (
            ru_maxrss_mb, settings.DIE_ON_RESIDENT_SET_SIZE_MB))

    def _on_sigtstp(self, num_completed, num_incomplete):
        """Callback function when SIGTSTP is triggered.

        :param int num_completed: the number of tasks completed in this batch.
        :param int num_incomplete: the number of tasks incomplete in the batch.
        """
        logger.info(
            ('Taskworker process suspended. Completed tasks: %d;'
             ' Incomplete: %d') % (
                num_completed, num_incomplete))

    def _on_shutdown(self, num_completed, num_incomplete):
        """Callback function when we shutdown the worker process.

        :param int num_completed: the number of tasks completed in this batch.
        :param int num_incomplete: the number of tasks incomplete in the batch.
        """
        logger.info(('Taskworker process shutting down. Completed tasks: %d;'
                     ' Incomplete: %d') % (num_completed, num_incomplete))

    def _on_pre_batch_run(self, messages):
        """Callback function before running a batch of tasks.

        :param list[KaleMessage] messages: a list of task messages to process.
        """
        logger.debug('Start processing %d tasks in a batch in queue %s ...' % (
            len(messages), self._batch_queue.name))

    def _on_post_batch_run(self, num_completed, num_incomplete, messages):
        """Callback function after running a batch of tasks.

        :param int num_completed: the number of tasks completed in this batch.
        :param int num_incomplete: the number of tasks incomplete in the batch.
        :param list[KaleMessage] messages: a list of messages in this batch.
        """
        logger.debug(('Finish processing message batch. Completed tasks: %d;'
                     ' Incomplete: %d') % (
            num_completed, num_incomplete))

    def _on_permanent_failure_batch(self):
        """Callback when there are permanently failed tasks in this batch."""
        logger.info(('Moving permamently %d failed tasks to the '
                     'dead-letter-queue %s.') % (
            len(self._permanent_failures), self._batch_queue.dlq_name))

    def _on_task_deferred(self, message, time_remaining_sec):
        """Callback function when a task is deferred.

        :param list[KaleMessage] message: a list of task messages.
        :param int time_remaining_sec: integer for seconds remained in the
            budget of running this batch of tasks.
        """
        task = message.task_inst
        logger.info(('Task deferred. Task id: %s; Queue: %s; '
                     'Time remaining: %d sec') % (
            task.task_id, self._batch_queue.name, time_remaining_sec))

    def _on_task_failed(self, message, time_remaining_sec, err,
                        permanent_failure):
        """Callback function when a task fails.

        :param KaleMessage message: an object of kale.message.KaleMessage.
        :param int time_remaining_sec: integer for seconds remained in the
            budget of running this batch of tasks.
        :param Exception err: object of Exception.
        :param bool permanent_failure: whether this task permanently fails.
        """
        task = message.task_inst
        logger.debug(('Task failed. Task id: %s; Queue: %s; '
                     'Time remaining: %d sec') % (
            task.task_id, self._batch_queue.name, time_remaining_sec))

    def _on_task_succeeded(self, message, time_remaining_sec):
        """Callback function when a task succeeds.

        :param KaleMessage message: an object of kale.message.KaleMessage.
        :param int time_remaining_sec: integer for seconds remained in the
            budget of running this batch of tasks.
        """
        task = message.task_inst
        logger.debug(('Task succeeded. Task id: %s; Queue: %s; '
                     'Time remaining: %d sec') % (
            task.task_id, self._batch_queue.name, time_remaining_sec))

    def run(self):
        """This method starts the task processing loop."""

        self._on_pre_run_worker()

        while self._check_process_resources():
            self._batch_queue = self._queue_selector.get_queue()
            for i in range(self._batch_queue.num_iterations):
                if not self._run_single_iteration():
                    # If the iteration didn't process any tasks break out
                    # of this loop and move on to another queue.
                    break

    def _check_process_resources(self):
        """Check if the process is still is abusing resources & should continue

        This will check the processes memory usage and gracefully exit if
        it exceeds the maximum value in settings.

        :return: True if still operational, otherwise it will exit the process.
        :rtype: bool
        """
        ru_maxrss_mb = utils.ru_maxrss_mb()

        if ru_maxrss_mb < settings.DIE_ON_RESIDENT_SET_SIZE_MB:

            if self._dirty:
                # We only log when the worker has been infected by  tasks.
                logger.debug('Worker process data.')
            return True

        # Allow the client of this library to do any setup before
        # shutting down the worker.
        settings.ON_WORKER_SHUTDOWN()

        self._on_exceeding_memory_limit(ru_maxrss_mb)

        # Use non-zero exit code.
        sys.exit(1)

    def _cleanup_worker(self, signum, frame):
        """Handle cleanup when the process is sent a signal.

        This will handle releasing tasks in flight and deleting tasks that have
        been completed.
        """
        logger.info('Process sent signal %d. Cleaning up tasks...' % signum)

        num_completed, num_incomplete = self._release_batch()

        # When the process is suspended we release tasks and then return to the
        # main loop.
        if signum == signal.SIGTSTP:
            self._on_sigtstp(num_completed, num_incomplete)
            return
        else:
            # Allow the client of this library to do any setup before
            # shutting down the worker.
            settings.ON_WORKER_SHUTDOWN()
            self._on_shutdown(num_completed, num_incomplete)
            sys.exit(0)

    def _run_single_iteration(self):
        """Run a single iteration of the task processing loop.

        :return: True if we were able to process a batch, False is there were
            no messages.
        """
        message_batch = self._consumer.fetch_batch(
            self._batch_queue.name,
            self._batch_queue.batch_size,
            self._batch_queue.visibility_timeout_sec,
            long_poll_time_sec=self._batch_queue.long_poll_time_sec)

        self._dirty = bool(message_batch)
        if not message_batch:
            # No any messages in this queue. Let's re-select queue again.
            return False

        self._on_pre_batch_run(message_batch)

        self._batch_stop_time = time.time() + \
            self._batch_queue.visibility_timeout_sec

        self._run_batch(message_batch)
        num_completed, num_incomplete = self._release_batch()

        self._on_post_batch_run(num_completed, num_incomplete, message_batch)

        return True

    def _release_batch(self):
        """Release the most recent batch back to SQS.
        This will delete tasks that succeeded and reset the
        visibility timeout for incomplete tasks.

        :return: A two-tuple of the count of tasks that were completed and
            the count of tasks that were incomplete.
        :rtype: tuple
        """
        # Delete from queues (failed messages are re-published as new tasks)
        messages_to_be_deleted = self._successful_messages + \
            self._failed_messages
        # Set timeout to 0 (if there is time left)
        # As an enhancement, we reset the timeout on tasks that didn't get
        # attempted if the remaining timeout is greater than some threshold.
        # and example of this being helpful is if a 5 minute task was given the
        # opportunity to run with 4 minutes left (and declines). This will
        # release the task 4 minutes before it previously would have.

        if (self._batch_stop_time - time.time()) > \
                settings.RESET_TIMEOUT_THRESHOLD:
            messages_to_be_released = self._incomplete_messages
        else:
            messages_to_be_released = []

        if messages_to_be_deleted:
            # Note: This includes failed tasks.
            self._consumer.delete_messages(messages_to_be_deleted,
                                           self._batch_queue.name)

        if messages_to_be_released:
            # This is only tasks that we didn't get the chance to attempt.
            self._consumer.release_messages(messages_to_be_released,
                                            self._batch_queue.name)

        # Send permanently failed tasks to the dead-letter-queue.
        if self._permanent_failures and settings.ENABLE_DEAD_LETTER_QUEUE:
            self._publisher.publish_messages_to_dead_letter_queue(
                self._batch_queue.dlq_name, self._permanent_failures)
            self._on_permanent_failure_batch()

        # All messages start as incomplete.
        self._incomplete_messages = []
        self._successful_messages = []
        self._failed_messages = []
        self._permanent_failures = []

        return len(messages_to_be_deleted), len(messages_to_be_released)

    def _run_batch(self, message_batch):
        """Consume as many tasks as possible in visibility timeout.

        :param list[KaleMessage] message_batch: list of consumable messages.
        """

        # All messages start as incomplete.
        self._incomplete_messages = list(message_batch)
        self._successful_messages = []
        self._failed_messages = []

        # These messages will be sent to the dead-letter-queue.
        self._permanent_failures = []

        # Set visibility timeout start time.
        for message in message_batch:
            time_remaining_sec = self._batch_stop_time - time.time()

            if message.task_inst.time_limit >= time_remaining_sec:
                # Greedily continue to process tasks, this task
                # is already in the incomplete_messages list
                self._on_task_deferred(message, time_remaining_sec)
                continue

            # Add cleanup method when tasks are timed out?

            self._run_single_message(message, time_remaining_sec)

    def _run_single_message(self, message, time_remaining_sec):
        """Call run_task on a single message and handle failures.

        This method handles time limiting, and calls subsequent commands if the task succeeds/fails

        :param message: KaleMessage
        :param time_remaining_sec: int seconds left for the batch of messages. Used mainly for logs
        """
        task_inst = message.task_inst
        try:
            with timeout.time_limit(task_inst.time_limit):
                self.run_task(message)
        except Exception as err:
            # Re-publish failed tasks.
            # As an optimization we could run all of the failures from a
            # batch together.
            permanent_failure = not task_inst.__class__.handle_failure(
                message, err)
            if permanent_failure and settings.USE_DEAD_LETTER_QUEUE:
                self._permanent_failures.append(message)

            self._failed_messages.append(message)

            self._on_task_failed(message, time_remaining_sec, err,
                                 permanent_failure)
        else:
            self._successful_messages.append(message)
            self._on_task_succeeded(message, time_remaining_sec)
        finally:
            self.remove_message_or_exit(message)
        # Increment total messages counter.
        self._total_messages_processed += 1

    def remove_message_or_exit(self, message):
        try:
            self._incomplete_messages.remove(message)
        except ValueError:
            # Cleanup happened due to the signal handler - make sure we exit immediately.
            sys.exit(0)

    def run_task(self, message):
        """Run the task contained in the message.
        :param message: message.KaleMessage containing the task and arguments to run.
        """
        message.task_inst.run(*message.task_args, **message.task_kwargs)
